from typing import Dict, List, Optional

import pytest


torch = pytest.importorskip("torch")
metatomic = pytest.importorskip("metatomic")

from metatensor.torch import Labels, TensorMap  # noqa: E402
from metatomic.torch import (  # noqa: E402
    AtomisticModel,
    ModelCapabilities,
    ModelMetadata,
    ModelOutput,
    NeighborListOptions,
    System,
)

from vesin.metatomic import (  # noqa: E402
    NeighborList,
    compute_requested_neighbors,
    compute_requested_neighbors_from_options,
)


def test_backward():
    positions = torch.tensor(
        [[0.0, 0.0, 0.0], [1.0, 1.0, 2.0]], dtype=torch.float64, requires_grad=True
    )
    cell = (4 * torch.eye(3, dtype=torch.float64)).clone().requires_grad_(True)
    system = System(
        positions=positions,
        cell=cell,
        pbc=torch.ones(3, dtype=bool),
        types=torch.tensor([6, 8]),
    )

    options = NeighborListOptions(cutoff=3.5, full_list=True, strict=True)
    calculator = NeighborList(options, length_unit="A")
    neighbors = calculator.compute(system)

    value = ((neighbors.values) ** 2).sum() * torch.linalg.det(cell)
    value.backward()

    # check there are gradients, and they are not zero
    assert positions.grad is not None
    assert cell.grad is not None
    assert torch.linalg.norm(positions.grad) > 0
    assert torch.linalg.norm(cell.grad) > 0


class InnerModule(torch.nn.Module):
    def requested_neighbor_lists(self) -> List[NeighborListOptions]:
        return [NeighborListOptions(cutoff=3.4, full_list=False, strict=True)]


class OuterModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.inner = InnerModule()

    def requested_neighbor_lists(self) -> List[NeighborListOptions]:
        return [NeighborListOptions(cutoff=5.2, full_list=True, strict=False)]

    def forward(
        self,
        systems: List[System],
        outputs: Dict[str, ModelOutput],
        selected_atoms: Optional[Labels],
    ) -> Dict[str, TensorMap]:
        return {}


class NLModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.inner = InnerModule()

    def requested_neighbor_lists(self) -> List[NeighborListOptions]:
        return [NeighborListOptions(cutoff=5.2, full_list=True, strict=False)]

    def forward(
        self,
        systems: List[System],
        outputs: Dict[str, ModelOutput],
        selected_atoms: Optional[Labels],
    ) -> None:
        compute_requested_neighbors_from_options(
            systems=systems,
            options=self.requested_neighbor_lists(),
            system_length_unit="A",
            check_consistency=True,
        )


def test_model():
    positions = torch.tensor(
        [[0.0, 0.0, 0.0], [1.0, 1.0, 1.4]], dtype=torch.float64, requires_grad=True
    )
    cell = (4 * torch.eye(3, dtype=torch.float64)).clone().requires_grad_(True)
    pbc = torch.ones(3, dtype=bool)
    types = torch.tensor([6, 8])
    systems = [
        System(positions=positions, cell=cell, pbc=pbc, types=types),
        System(positions=positions, cell=cell, pbc=pbc, types=types),
    ]

    # Using a "raw" model
    model = OuterModule()
    compute_requested_neighbors(
        systems=systems,
        system_length_unit="A",
        model=model,
        model_length_unit="A",
        check_consistency=True,
    )

    for system in systems:
        all_options = system.known_neighbor_lists()
        assert len(all_options) == 2
        assert all_options[0].requestors() == ["OuterModule"]
        assert all_options[0].cutoff == 5.2
        assert all_options[1].requestors() == ["OuterModule.inner"]
        assert all_options[1].cutoff == 3.4

    # Using a AtomisticModel
    capabilities = ModelCapabilities(
        length_unit="A",
        interaction_range=6.0,
        supported_devices=["cpu"],
        dtype="float64",
    )
    model = AtomisticModel(model.eval(), ModelMetadata(), capabilities)
    compute_requested_neighbors(
        systems=System(positions=positions, cell=cell, pbc=pbc, types=types),
        system_length_unit="A",
        model=model,
    )

    for system in systems:
        all_options = system.known_neighbor_lists()
        assert len(all_options) == 2
        assert all_options[0].requestors() == ["OuterModule"]
        assert all_options[0].cutoff == 5.2
        assert all_options[1].requestors() == ["OuterModule.inner"]
        assert all_options[1].cutoff == 3.4

    message = (
        "the given `model_length_unit` \\(nm\\) does not match the model "
        "capabilities \\(A\\)"
    )
    with pytest.raises(ValueError, match=message):
        compute_requested_neighbors(
            systems=System(positions=positions, cell=cell, pbc=pbc, types=types),
            system_length_unit="A",
            model=model,
            model_length_unit="nm",
        )


def test_torchscriptability():
    torch.jit.script(NLModule())



import itertools
import os

import ase.io
import ase.neighborlist
import numpy as np
import pytest

from vesin import NeighborList


cp = pytest.importorskip("cupy")

# Check if CUDA is available
try:
    cp.cuda.Device(0).compute_capability
    # all good
except cp.cuda.runtime.CUDARuntimeError:
    pytest.skip("CUDA is not available", allow_module_level=True)


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))


@pytest.mark.parametrize("full_list", [False, True])
def test_large_box_small_cutoff(full_list):
    points = cp.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 2.0, 0.0],
            [0.0, 0.0, 2.0],
            [-6.0, 0.0, 0.0],
            [-6.0, -2.0, 0.0],
            [-6.0, 0.0, -2.0],
        ],
        dtype=cp.float64,
    )

    box = cp.array(
        [
            [54.0, 0.0, 0.0],
            [0.0, 54.0, 0.0],
            [0.0, 0.0, 54.0],
        ],
        dtype=cp.float64,
    )

    calculator = NeighborList(cutoff=2.1, full_list=full_list)

    i, j, S, d, D = calculator.compute(points, box, periodic=True, quantities="ijSdD")

    # check all outputs are cupy arrays
    assert isinstance(i, cp.ndarray)
    assert isinstance(j, cp.ndarray)
    assert isinstance(S, cp.ndarray)
    assert isinstance(d, cp.ndarray)
    assert isinstance(D, cp.ndarray)

    assert len(i) == len(j)
    assert len(i) == len(d)
    assert len(i) == len(D)
    assert len(i) == len(S)
    assert len(i) == (8 if full_list else 4)

    pairs = cp.stack((i, j), axis=1)
    sort_idx = cp.argsort(pairs[:, 0] * (i.max() + 1) + pairs[:, 1])

    # Apply sort
    i = i[sort_idx]
    j = j[sort_idx]
    S = S[sort_idx]
    d = d[sort_idx]
    D = D[sort_idx]

    # Convert to plain Python lists for easy matching
    actual_pairs = sorted(zip(i.tolist(), j.tolist(), strict=True))

    if full_list:
        expected_pairs = [
            (0, 1),
            (0, 2),
            (1, 0),
            (2, 0),
            (3, 4),
            (3, 5),
            (4, 3),
            (5, 3),
        ]

        expected_shifts = cp.zeros((8, 3), dtype=cp.int32)
        expected_distances = cp.array(
            [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
            dtype=cp.float64,
        )
        expected_vectors = cp.array(
            [
                [0.0, 2.0, 0.0],
                [0.0, 0.0, 2.0],
                [0.0, -2.0, 0.0],
                [0.0, 0.0, -2.0],
                [0.0, -2.0, 0.0],
                [0.0, 0.0, -2.0],
                [0.0, 2.0, 0.0],
                [0.0, 0.0, 2.0],
            ],
            dtype=cp.float64,
        )
    else:
        expected_pairs = sorted([(0, 1), (0, 2), (3, 4), (3, 5)])
        expected_shifts = cp.zeros((4, 3), dtype=cp.int32)
        expected_distances = cp.array([2.0, 2.0, 2.0, 2.0], dtype=cp.float64)
        expected_vectors = cp.array(
            [
                [0.0, 2.0, 0.0],
                [0.0, 0.0, 2.0],
                [0.0, -2.0, 0.0],
                [0.0, 0.0, -2.0],
            ],
            dtype=cp.float64,
        )

    assert actual_pairs == expected_pairs
    assert cp.all(S == expected_shifts)
    assert cp.allclose(d, expected_distances)
    assert cp.allclose(D, expected_vectors)


@pytest.mark.parametrize("system", ["water", "diamond", "naphthalene", "carbon"])
@pytest.mark.parametrize("algorithm", ["brute_force", "cell_list", "auto"])
def test_neighbors(system, algorithm):
    atoms = ase.io.read(f"{CURRENT_DIR}/../../vesin/tests/data/{system}.xyz")

    # make the cell bigger for MIC
    if not np.allclose(atoms.cell, np.zeros((3, 3))):
        atoms = atoms.repeat((2, 2, 2))

    cutoff = 2.0

    ase_i, ase_j, ase_S, ase_D = ase.neighborlist.neighbor_list("ijSD", atoms, cutoff)

    calculator = NeighborList(cutoff=cutoff, full_list=True, algorithm=algorithm)
    vesin_i, vesin_j, vesin_S, vesin_D = calculator.compute(
        points=cp.array(atoms.positions),
        box=cp.array(atoms.cell[:]),
        periodic=cp.array(atoms.pbc),
        quantities="ijSD",
    )

    # get as numpy arrays
    vesin_i = cp.asnumpy(vesin_i)
    vesin_j = cp.asnumpy(vesin_j)
    vesin_S = cp.asnumpy(vesin_S)
    vesin_D = cp.asnumpy(vesin_D)

    assert len(ase_i) == len(vesin_i)
    assert len(ase_j) == len(vesin_j)
    assert len(ase_S) == len(vesin_S)
    assert len(ase_D) == len(vesin_D)

    ase_ijS = np.concatenate(
        (ase_i.reshape(-1, 1), ase_j.reshape(-1, 1), ase_S), axis=1
    )
    vesin_ijS = np.concatenate(
        (vesin_i.reshape(-1, 1), vesin_j.reshape(-1, 1), vesin_S), axis=1
    )

    ase_sort_indices = np.lexsort(np.flip(ase_ijS, axis=1).T)
    vesin_sort_indices = np.lexsort(np.flip(vesin_ijS, axis=1).T)

    assert np.array_equal(ase_ijS[ase_sort_indices], vesin_ijS[vesin_sort_indices])
    assert np.allclose(ase_D[ase_sort_indices], vesin_D[vesin_sort_indices])


@pytest.mark.parametrize(
    "periodic",
    list(itertools.product([False, True], repeat=3)),
)
@pytest.mark.parametrize("algorithm", ["brute_force", "cell_list", "auto"])
def test_mixed_periodic(periodic, algorithm):
    cutoff = 0.35
    # Use a fixed seed for the box to ensure reproducibility
    rng = np.random.default_rng(42)
    box = np.eye(3, dtype=np.float64)[[2, 0, 1]] + 0.1 * rng.normal(size=(3, 3))
    points = np.random.default_rng(0).random((100, 3))

    atoms = ase.Atoms(positions=points, cell=box, pbc=periodic)
    ase_i, ase_j, ase_S, ase_D, ase_d = ase.neighborlist.neighbor_list(
        "ijSDd", atoms, cutoff
    )

    calculator = NeighborList(cutoff=cutoff, full_list=True, algorithm=algorithm)
    vesin_i, vesin_j, vesin_S, vesin_D, vesin_d = calculator.compute(
        points=cp.array(points, dtype=cp.float64),
        box=cp.array(box, dtype=cp.float64),
        periodic=cp.array(periodic),
        quantities="ijSDd",
    )

    vesin_i = cp.asnumpy(vesin_i)
    vesin_j = cp.asnumpy(vesin_j)
    vesin_S = cp.asnumpy(vesin_S)
    vesin_D = cp.asnumpy(vesin_D)
    vesin_d = cp.asnumpy(vesin_d)

    assert len(ase_i) == len(vesin_i)
    assert len(ase_j) == len(vesin_j)
    assert len(ase_S) == len(vesin_S)
    assert len(ase_D) == len(vesin_D)
    assert len(ase_d) == len(vesin_d)

    ase_ijS = np.concatenate(
        (ase_i.reshape(-1, 1), ase_j.reshape(-1, 1), ase_S), axis=1
    )
    vesin_ijS = np.concatenate(
        (vesin_i.reshape(-1, 1), vesin_j.reshape(-1, 1), vesin_S), axis=1
    )

    ase_sort_indices = np.lexsort(np.flip(ase_ijS, axis=1).T)
    vesin_sort_indices = np.lexsort(np.flip(vesin_ijS, axis=1).T)

    assert np.array_equal(ase_ijS[ase_sort_indices], vesin_ijS[vesin_sort_indices])
    assert np.allclose(ase_D[ase_sort_indices], vesin_D[vesin_sort_indices])
    assert np.allclose(ase_d[ase_sort_indices], vesin_d[vesin_sort_indices])


def test_no_neighbors():
    """Test CUDA implementation when there are no neighbors"""
    points = cp.array([[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]], dtype=cp.float64)
    box = cp.eye(3, dtype=cp.float64)

    calculator = NeighborList(cutoff=0.1, full_list=True)
    i, j = calculator.compute(points, box, True, quantities="ij")

    assert len(i) == 0
    assert len(j) == 0


@pytest.mark.parametrize("dtype", [cp.float32, cp.float64])
def test_dtype(dtype):
    box = cp.eye(3, dtype=dtype) * 3.0
    points = cp.array(np.random.default_rng(0).random((100, 3), dtype=dtype) * 3.0)

    calculator = NeighborList(cutoff=1, full_list=True)
    i, j, s, D, d = calculator.compute(points, box, True, "ijSDd")

    assert i.dtype == cp.uint64
    assert j.dtype == cp.uint64
    assert s.dtype == cp.int32
    assert D.dtype == dtype
    assert d.dtype == dtype


def test_max_pairs():
    """
    With the default VESIN_CUDA_MAX_PAIRS_PER_POINT=512 and 100 points,
    a cutoff of 4.0 generates too many pairs and should raise an error.
    """
    dtype = cp.float64

    box = cp.eye(3, dtype=dtype) * 3.0
    points = cp.array(np.random.default_rng(0).random((100, 3), dtype=dtype) * 3.0)

    message = (
        "The number of neighbor pairs exceeds the maximum capacity of 30720 "
        "\\(max_pairs_per_point=256; n_points=100\\). Consider reducing the "
        "cutoff distance, or explicitly setting VESIN_CUDA_MAX_PAIRS_PER_POINT as "
        "an environment variable."
    )
    calculator = NeighborList(cutoff=4, full_list=True)
    with pytest.raises(RuntimeError, match=message):
        _ = calculator.compute(points, box, True, "ijSDd")

    os.environ["VESIN_CUDA_MAX_PAIRS_PER_POINT"] = "1000"
    # This should now work without error
    calculator = NeighborList(cutoff=4, full_list=True)
    _ = calculator.compute(points, box, True, "ijSDd")

    os.environ["VESIN_CUDA_MAX_PAIRS_PER_POINT"] = "10"
    message = (
        "The number of neighbor pairs exceeds the maximum capacity of 1200 "
        "\\(max_pairs_per_point=10; n_points=100\\). Consider reducing the "
        "cutoff distance, or explicitly setting VESIN_CUDA_MAX_PAIRS_PER_POINT as "
        "an environment variable."
    )
    # error message should show the new value of VESIN_CUDA_MAX_PAIRS_PER_POINT
    calculator = NeighborList(cutoff=4, full_list=True)
    with pytest.raises(RuntimeError, match=message):
        _ = calculator.compute(points, box, True, "ijSDd")

    for invalid in ["-3", "44 ABC", "0", "-1", "1.5", ""]:
        os.environ["VESIN_CUDA_MAX_PAIRS_PER_POINT"] = invalid
        message = f"Invalid value for VESIN_CUDA_MAX_PAIRS_PER_POINT: '{invalid}'"
        calculator = NeighborList(cutoff=4, full_list=True)
        with pytest.raises(RuntimeError, match=message):
            _ = calculator.compute(points, box, True, "ijSDd")

    del os.environ["VESIN_CUDA_MAX_PAIRS_PER_POINT"]



from typing import List, Union

import torch
from metatensor.torch import Labels, TensorBlock
from metatomic.torch import NeighborListOptions, System, register_autograd_neighbors

from .. import NeighborList as NeighborListNumpy


try:
    from vesin.torch import NeighborList as NeighborListTorch

except ImportError:

    class NeighborListTorch:
        def __init__(self, cutoff: float, full_list: bool):
            raise ValueError("torchscript=True requires `vesin-torch` as a dependency")

        def compute(
            self,
            points: torch.Tensor,
            box: torch.Tensor,
            periodic: Union[bool, torch.Tensor],
            quantities: str,
            copy: bool = True,
        ) -> List[torch.Tensor]:
            raise ValueError("torchscript=True requires `vesin-torch` as a dependency")


class NeighborList:
    """
    A neighbor list calculator that can be used with metatomic's models.

    The main difference with the other calculators is the automatic handling of
    different length unit between what the model expects and what the ``System`` are
    using.

    .. seealso::

        The :py:func:`vesin.metatomic.compute_requested_neighbors` function can be used
        to automatically compute and store all neighbor lists required by a given model.
    """

    def __init__(
        self,
        options: NeighborListOptions,
        length_unit: str,
        torchscript: bool = False,
        check_consistency: bool = False,
    ):
        """
        :param options: :py:class:`metatomic.torch.NeighborListOptions` defining the
            parameters of the neighbor list
        :param length_unit: unit of length used for the systems data
        :param torchscript: whether this function should be compatible with TorchScript
            or not. If ``True``, this requires installing the ``vesin-torch`` package.
        :param check_consistency: whether to run additional checks on the neighbor list
            validity

        Example
        -------

        >>> from vesin.metatomic import NeighborList
        >>> from metatomic.torch import System, NeighborListOptions
        >>> import torch
        >>> system = System(
        ...     positions=torch.eye(3).requires_grad_(True),
        ...     cell=4 * torch.eye(3).requires_grad_(True),
        ...     types=torch.tensor([8, 1, 1]),
        ...     pbc=torch.ones(3, dtype=bool),
        ... )
        >>> options = NeighborListOptions(cutoff=4.0, full_list=True, strict=False)
        >>> calculator = NeighborList(options, length_unit="Angstrom")
        >>> neighbors = calculator.compute(system)
        >>> neighbors
        TensorBlock
            samples (18): ['first_atom', 'second_atom', 'cell_shift_a', 'cell_shift_b', 'cell_shift_c']
            components (3): ['xyz']
            properties (1): ['distance']
            gradients: None
        <BLANKLINE>

        The returned TensorBlock can then be registered with the system

        >>> system.add_neighbor_list(options, neighbors)
        """  # noqa: E501

        self.options = options
        self.length_unit = length_unit
        self.check_consistency = check_consistency

        if torch.jit.is_scripting() or torchscript:
            self._nl = NeighborListTorch(
                cutoff=self.options.engine_cutoff(self.length_unit),
                full_list=self.options.full_list,
            )
        else:
            self._nl = NeighborListNumpy(
                cutoff=self.options.engine_cutoff(self.length_unit),
                full_list=self.options.full_list,
            )

        # cached Labels
        self._components = Labels("xyz", torch.tensor([[0], [1], [2]]))
        self._properties = Labels("distance", torch.tensor([[0]]))

    def compute(self, system: System) -> TensorBlock:
        """
        Compute the neighbor list for the given :py:class:`metatomic.torch.System`.

        :param system: a :py:class:`metatomic.torch.System` containing data about a
            single structure. If the positions or cell of this system require gradients,
            the neighbors list values computational graph will be set accordingly.

            The positions and cell need to be in the length unit defined for this
            :py:class:`NeighborList` calculator.
        """

        points = system.positions.detach()
        box = system.cell.detach()

        # computes neighbor list
        (P, S, D) = self._nl.compute(
            points=points, box=box, periodic=system.pbc, quantities="PSD", copy=True
        )
        P = torch.as_tensor(P, dtype=torch.int32)
        S = torch.as_tensor(S, dtype=torch.int32)
        D = torch.as_tensor(D)

        self._components = self._components.to(device=D.device)
        self._properties = self._properties.to(device=D.device)

        # converts to a suitable TensorBlock format
        neighbors = TensorBlock(
            D.reshape(-1, 3, 1),
            samples=Labels(
                names=[
                    "first_atom",
                    "second_atom",
                    "cell_shift_a",
                    "cell_shift_b",
                    "cell_shift_c",
                ],
                values=torch.hstack([P, S]),
            ),
            components=[self._components],
            properties=self._properties,
        )

        register_autograd_neighbors(
            system, neighbors, check_consistency=self.check_consistency
        )

        return neighbors



import warnings

import metatomic.torch
import torch


# Dynamically check the dependencies versions since we can not declare them
# as dependencies in pyproject.toml (we only want to provide code in this module
# if the user already has the correct dependencies installed).
torch_version = tuple(map(int, torch.__version__.split(".")[:2]))
if torch_version < (2, 3):
    raise ImportError(
        f"Found torch v{torch.__version__}, but vesin.metatomic requires torch >= 2.3"
    )

mta_version = tuple(map(int, metatomic.torch.__version__.split(".")[:3]))
if mta_version < (0, 1, 3) or mta_version >= (0, 2, 0):
    # this is not an import error so we can also use this code inside metatomic
    # to implement interfaces to some Python simulation engines and not fail when
    # building development versions of metatomic.torch
    warnings.warn(
        f"Found metatomic.torch v{metatomic.torch.__version__}, but vesin.metatomic "
        "was only tested with metatomic.torch >=0.1.3,<0.2. This may lead to "
        "unexpected errors. Please install a compatible version of metatomic.torch if "
        "you encounter any issues.",
        stacklevel=1,
    )


from ._model import (  # noqa: E402
    compute_requested_neighbors,
    compute_requested_neighbors_from_options,
)
from ._neighbors import NeighborList  # noqa: E402


__all__ = [
    "NeighborList",
    "compute_requested_neighbors",
    "compute_requested_neighbors_from_options",
]



import os
import re
import shutil
import subprocess
import sys
import urllib.request

from setuptools import Extension, setup
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.bdist_wheel import bdist_wheel
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist


ROOT = os.path.realpath(os.path.dirname(__file__))

VESIN_BUILD_TYPE = os.environ.get("VESIN_BUILD_TYPE", "Release")
if VESIN_BUILD_TYPE not in ["Debug", "Release"]:
    raise Exception(
        f"invalid build type passed: '{VESIN_BUILD_TYPE}', "
        "expected 'Debug' or 'Release'"
    )


class universal_wheel(bdist_wheel):
    # When building the wheel, the `wheel` package assumes that if we have a
    # binary extension then we are linking to `libpython.so`; and thus the wheel
    # is only usable with a single python version. This is not the case for
    # here, and the wheel will be compatible with any Python >=3.7. This is
    # tracked in https://github.com/pypa/wheel/issues/185, but until then we
    # manually override the wheel tag.
    def get_tag(self):
        tag = bdist_wheel.get_tag(self)
        # tag[2:] contains the os/arch tags, we want to keep them
        return ("py3", "none") + tag[2:]


class cmake_ext(build_ext):
    """
    Build the native library using cmake
    """

    def run(self):
        source_dir = os.path.join(ROOT, "lib")
        if not os.path.exists(source_dir):
            # we are building from a checkout
            source_dir = os.path.join(ROOT, "..", "..", "vesin")

        build_dir = os.path.join(ROOT, "build", "cmake-build")
        install_dir = os.path.join(os.path.realpath(self.build_lib), "vesin")

        os.makedirs(build_dir, exist_ok=True)

        cmake_options = [
            f"-DCMAKE_INSTALL_PREFIX={install_dir}",
            f"-DCMAKE_BUILD_TYPE={VESIN_BUILD_TYPE}",
            "-DBUILD_SHARED_LIBS=ON",
        ]

        CUDA_HOME = os.environ.get("CUDA_HOME")
        VESIN_ENABLE_NVTX = os.environ.get("VESIN_ENABLE_NVTX", "OFF")

        if CUDA_HOME is not None:
            cmake_options.append(f"-DCUDA_TOOLKIT_ROOT_DIR={CUDA_HOME}")
            cmake_options.append("-DVESIN_ENABLE_CUDA=ON")
            if VESIN_ENABLE_NVTX.upper() in ["ON", "1", "TRUE", "YES"]:
                cmake_options.append("-DVESIN_ENABLE_NVTX=ON")

            # fix for https://github.com/pytorch/pytorch/issues/113948, it does not
            # matter which architecture we put here since we are not actually compiling
            # any CUDA code, just linking against cudart
            cmake_options.append("-DTORCH_CUDA_ARCH_LIST=9.0")

        subprocess.run(
            ["cmake", source_dir, *cmake_options],
            cwd=build_dir,
            check=True,
        )

        subprocess.run(
            [
                "cmake",
                "--build",
                build_dir,
                "--config",
                VESIN_BUILD_TYPE,
                "--target",
                "install",
            ],
            check=True,
        )


class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg

    Prevents setup.py install performing setuptools' default easy_install,
    which it should never ever do.
    """

    def run(self):
        sys.exit(
            "Aborting implicit building of eggs.\nUse `pip install .` or "
            "`python -m build --wheel . && pip install "
            "dist/metatensor_core-*.whl` to install from source."
        )


class sdist_with_lib(sdist):
    """
    Create a sdist including the code for the native library
    """

    def run(self):
        # generate extra files
        shutil.copytree(os.path.join(ROOT, "..", "..", "vesin"), "lib")

        # include gpulite in the sdist
        gpulite_dir = os.path.join("lib", "external")
        os.makedirs(gpulite_dir, exist_ok=True)
        gpulite_archive = os.path.join(gpulite_dir, "gpulite.tar.gz")
        assert not os.path.exists(gpulite_archive)

        with open(os.path.join("lib", "CMakeLists.txt")) as fd:
            content = fd.read()
            # FetchContent_Declare(
            #     gpulite
            #     ...
            #     GIT_TAG <hash>
            match = re.search(
                r"FetchContent_Declare\s*\(\s*gpulite.*?GIT_TAG\s+([a-f0-9]+)",
                content,
                re.DOTALL,
            )
            if match is None:
                raise Exception("Could not find gpulite GIT_TAG in CMakeLists.txt")
            commit = match.group(1)

        print("downloading gpulite source code")
        urllib.request.urlretrieve(
            f"https://github.com/rubber-duck-debug/gpu-lite/archive/{commit}.tar.gz",
            gpulite_archive,
        )

        # run original sdist
        super().run()

        # cleanup
        shutil.rmtree("lib")


if __name__ == "__main__":
    setup(
        version=open("VERSION").read().strip(),
        ext_modules=[
            # only declare the extension, it is built & copied as required by cmake
            # in the build_ext command
            Extension(name="vesin", sources=[]),
        ],
        cmdclass={
            "sdist": sdist_with_lib,
            "build_ext": cmake_ext,
            "bdist_egg": bdist_egg if "bdist_egg" in sys.argv else bdist_egg_disabled,
            "bdist_wheel": universal_wheel,
        },
    )



import ctypes
import glob
import os
import sys
from ctypes import POINTER
from typing import List, Sequence, Union

import numpy as np
import numpy.typing as npt

from ._c_api import (
    VesinAutoAlgorithm,
    VesinBruteForce,
    VesinCellList,
    VesinCPU,
    VesinCUDA,
    VesinDevice,
    VesinNeighborList,
    VesinOptions,
    VesinUnknownDevice,
)
from ._c_lib import _get_library


try:
    import cupy as cp

    HAS_CUPY = True

    CUPY_DTYPES = {
        ctypes.c_size_t: cp.uint64,
        ctypes.c_int32: cp.int32,
        ctypes.c_double: cp.float64,
    }
except ImportError:
    HAS_CUPY = False

try:
    import torch

    HAS_TORCH = True

    TORCH_DTYPES = {
        ctypes.c_size_t: torch.uint64,
        ctypes.c_int32: torch.int32,
        ctypes.c_double: torch.float64,
    }
except ImportError:
    HAS_TORCH = False


def _device_from_array(array):
    """
    Determine the VesinDevice from a numpy or cupy array.

    :param array: numpy array or cupy array
    :return: VesinDevice structure
    """
    if HAS_CUPY and isinstance(array, cp.ndarray):
        return VesinDevice(VesinCUDA, array.device.id)
    elif HAS_TORCH and isinstance(array, torch.Tensor):
        if array.device.type == "cuda":
            return VesinDevice(VesinCUDA, array.device.index or 0)
        else:
            return VesinDevice(VesinCPU, 0)
    else:
        return VesinDevice(VesinCPU, 0)


class NeighborList:
    """
    A neighbor list calculator.
    """

    def __init__(
        self,
        cutoff: float,
        full_list: bool,
        sorted: bool = False,
        algorithm: str = "auto",
    ):
        """
        :param cutoff: spherical cutoff for this neighbor list
        :param full_list: should we return each pair twice (as ``i-j`` and ``j-i``) or
            only once
        :param sorted: Should vesin sort the returned pairs in lexicographic order
            (sorting both ``i`` and then ``j`` at constant ``i``)?
        :param algorithm: algorithm to use when computing the neighbor list. One of
            ``"auto"``, ``"brute_force"``, or ``"cell_list"``.
        """
        self._lib = _get_library()
        self.cutoff = float(cutoff)
        self.full_list = bool(full_list)
        self.sorted = bool(sorted)

        self.algorithm = algorithm

        self._neighbors = VesinNeighborList()

    def __del__(self):
        if hasattr(self, "_lib") and hasattr(self, "_neighbors"):
            self._lib.vesin_free(self._neighbors)

    @property
    def algorithm(self) -> str:
        """Get the current algorithm used for NL computation."""
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value: str):
        """Set the algorithm to use for NL computation."""
        algorithm = value.lower()
        self._algorithm = algorithm
        if algorithm == "auto":
            self._c_algorithm = VesinAutoAlgorithm
        elif algorithm == "brute_force":
            self._c_algorithm = VesinBruteForce
        elif algorithm == "cell_list":
            self._c_algorithm = VesinCellList
        else:
            raise ValueError(
                f"Unknown algorithm '{algorithm}'. "
                "Must be one of 'auto', 'brute_force', or 'cell_list'."
            )

    def compute(
        self,
        points: "npt.ArrayLike",
        box: "npt.ArrayLike",
        periodic: Union[bool, Sequence[bool], "npt.ArrayLike"],
        quantities: str = "ij",
        copy=True,
    ) -> List[np.ndarray]:
        """
        Compute the neighbor list for the system defined by ``positions``, ``box``, and
        ``periodic``; returning the requested ``quantities``.

        ``quantities`` can contain any combination of the following values:

        - ``"i"`` to get the index of the first point in the pair
        - ``"j"`` to get the index of the second point in the pair
        - ``"P"`` to get the indexes of the two points in the pair simultaneously
        - ``"S"`` to get the periodic shift of the pair
        - ``"d"`` to get the distance between points in the pair
        - ``"D"`` to get the distance vector between points in the pair

        :param points: positions of all points in the system (this can be anything that
            can be converted to a numpy or cupy array)
        :param box: bounding box of the system (this can be anything that can be
            converted to a numpy or cupy array)
        :param periodic: should we use periodic boundary conditions? This can be a
            single boolean to enable/disable periodic boundary conditions in all
            directions, or a sequence of three booleans (one for each direction).
        :param quantities: quantities to return, defaults to "ij"
        :param copy: should we copy the returned quantities, defaults to ``True``.
            Setting this to ``False`` might be a bit faster, but the returned arrays are
            view inside this class, and will be invalidated whenever this class is
            garbage collected or used to run a new calculation.

        :return: tuple of arrays as indicated by ``quantities``.
        """
        use_cupy = HAS_CUPY and isinstance(points, cp.ndarray)
        use_torch = HAS_TORCH and isinstance(points, torch.Tensor)

        if use_cupy:
            get_ptr_fn = _cupy_get_ptr
            as_array_fn = _cupy_asarray
            to_dtype_fn = _cupy_to_dtype
            float64_dtype = cp.float64
            bool_dtype = cp.bool_
            empty_array_fn = _cupy_empty
            ptr_to_array_fn = _ptr_to_cupy
            copy_array_fn = lambda arr: arr.copy()  # noqa: E731
        elif use_torch:
            get_ptr_fn = _torch_get_ptr
            as_array_fn = _torch_asarray
            to_dtype_fn = _torch_to_dtype
            float64_dtype = torch.float64
            bool_dtype = torch.bool
            empty_array_fn = _torch_empty
            ptr_to_array_fn = _ptr_to_torch
            copy_array_fn = lambda arr: arr.clone()  # noqa: E731
        else:
            # use numpy by default
            get_ptr_fn = _numpy_get_ptr
            as_array_fn = _numpy_asarray
            to_dtype_fn = _numpy_to_dtype
            float64_dtype = np.float64
            bool_dtype = np.bool_
            empty_array_fn = _numpy_empty
            ptr_to_array_fn = _ptr_to_numpy
            copy_array_fn = lambda arr: arr.copy()  # noqa: E731

        if box.shape != (3, 3):
            raise ValueError("`box` must be a 3x3 matrix")

        if len(points.shape) != 2 or points.shape[1] != 3:
            raise ValueError("`points` must be a nx3 array")

        options = VesinOptions()
        options.cutoff = self.cutoff
        options.full = self.full_list
        options.sorted = self.sorted
        options.return_shifts = "S" in quantities
        options.return_distances = "d" in quantities
        options.return_vectors = "D" in quantities
        options.algorithm = self._c_algorithm

        if isinstance(periodic, (bool, np.bool_)):
            periodic = np.array([periodic, periodic, periodic], dtype=np.bool_)

        # Get device and data pointer
        device = _device_from_array(points)

        points = as_array_fn(points, device_like=None)
        box = as_array_fn(box, device_like=None)
        periodic = as_array_fn(periodic, device_like=points)
        periodic = to_dtype_fn(periodic, bool_dtype)

        if periodic.shape != (3,):
            raise ValueError(
                "`periodic` must be a single boolean or a sequence of three "
                f"booleans, got {periodic} of type {type(periodic)}"
            )

        initial_dtype = points.dtype
        if box.dtype != initial_dtype:
            raise RuntimeError(
                "`points` and `box` must have the same dtype, "
                f"got {points.dtype} and {box.dtype}"
            )

        points = to_dtype_fn(points, float64_dtype)
        box = to_dtype_fn(box, float64_dtype)

        if use_torch:
            if points.requires_grad or box.requires_grad:
                raise RuntimeError(
                    "NeighborList.compute does not support autograd for torch tensors"
                )

        points, points_ptr = get_ptr_fn(points, ctypes.c_double)
        box, box_ptr = get_ptr_fn(box, ctypes.c_double)
        periodic, periodic_ptr = get_ptr_fn(periodic, ctypes.c_bool)

        if self._neighbors.device.type != VesinUnknownDevice:
            if (
                self._neighbors.device.type != device.type
                or self._neighbors.device.device_id != device.device_id
            ):
                # Free previous allocation and reset to zeroed state
                self._lib.vesin_free(self._neighbors)
                ctypes.memset(
                    ctypes.byref(self._neighbors),
                    0,
                    ctypes.sizeof(self._neighbors),
                )

        error_message = ctypes.c_char_p()
        status = self._lib.vesin_neighbors(
            points_ptr,
            points.shape[0],
            box_ptr,
            periodic_ptr,
            device,
            options,
            ctypes.byref(self._neighbors),
            error_message,
        )

        if status != 0:
            raise RuntimeError(error_message.value.decode("utf8"))

        # Create arrays for the output data
        n_pairs = self._neighbors.length

        if n_pairs == 0:
            # Empty results -- use initial_dtype for floating-point arrays so
            # the caller gets the same dtype they passed in.
            pairs = empty_array_fn((0, 2), dtype=ctypes.c_size_t, device_like=points)
            shifts = empty_array_fn((0, 3), dtype=ctypes.c_int32, device_like=points)
            distances = empty_array_fn((0,), dtype=initial_dtype, device_like=points)
            vectors = empty_array_fn((0, 3), dtype=initial_dtype, device_like=points)
        else:
            pairs = ptr_to_array_fn(
                self._neighbors.pairs,
                shape=(n_pairs, 2),
                dtype=ctypes.c_size_t,
                owner=self._neighbors,
                device=self._neighbors.device,
            )
            if "S" in quantities:
                shifts = ptr_to_array_fn(
                    self._neighbors.shifts,
                    shape=(n_pairs, 3),
                    dtype=ctypes.c_int32,
                    owner=self._neighbors,
                    device=self._neighbors.device,
                )
            if "d" in quantities:
                distances = ptr_to_array_fn(
                    self._neighbors.distances,
                    shape=(n_pairs,),
                    dtype=ctypes.c_double,
                    owner=self._neighbors,
                    device=self._neighbors.device,
                )
                distances = to_dtype_fn(distances, initial_dtype)
            if "D" in quantities:
                vectors = ptr_to_array_fn(
                    self._neighbors.vectors,
                    shape=(n_pairs, 3),
                    dtype=ctypes.c_double,
                    owner=self._neighbors,
                    device=self._neighbors.device,
                )
                vectors = to_dtype_fn(vectors, initial_dtype)

        # assemble output

        data = []
        for quantity in quantities:
            if quantity == "P":
                if copy:
                    data.append(copy_array_fn(pairs))
                else:
                    data.append(pairs)

            if quantity == "i":
                if copy:
                    data.append(copy_array_fn(pairs[:, 0]))
                else:
                    data.append(pairs[:, 0])

            elif quantity == "j":
                if copy:
                    data.append(copy_array_fn(pairs[:, 1]))
                else:
                    data.append(pairs[:, 1])

            elif quantity == "S":
                if copy:
                    data.append(copy_array_fn(shifts))
                else:
                    data.append(shifts)

            elif quantity == "d":
                if copy:
                    data.append(copy_array_fn(distances))
                else:
                    data.append(distances)

            elif quantity == "D":
                if copy:
                    data.append(copy_array_fn(vectors))
                else:
                    data.append(vectors)

        return tuple(data)


########################################################################################


def _numpy_asarray(array, device_like) -> np.ndarray:
    return np.asarray(array)


def _numpy_to_dtype(array: np.ndarray, dtype) -> np.ndarray:
    return array.astype(dtype)


def _numpy_empty(shape, dtype, device_like) -> np.ndarray:
    return np.empty(shape, dtype=dtype)


def _numpy_get_ptr(array: np.ndarray, dtype) -> (np.ndarray, POINTER):
    array = np.ascontiguousarray(array)
    return array, array.ctypes.data_as(POINTER(dtype))


def _ptr_to_numpy(ptr, shape, dtype, owner, device):
    """
    Helper to convert a ctypes pointer to a NumPy array.
    """
    assert device.type == VesinCPU, "Array is not on CPU"
    return np.ctypeslib.as_array(ctypes.cast(ptr, POINTER(dtype)), shape=shape)


########################################################################################


def _cupy_asarray(array, device_like) -> "cp.ndarray":
    return cp.asarray(array)


def _cupy_to_dtype(array: "cp.ndarray", dtype) -> "cp.ndarray":
    return array.astype(dtype)


def _cupy_empty(shape, dtype, device_like) -> "cp.ndarray":
    return cp.empty(shape, dtype=dtype)


def _cupy_get_ptr(array: "cp.ndarray", dtype) -> ("cp.ndarray", POINTER):
    array = cp.ascontiguousarray(array)
    ptr = ctypes.cast(ctypes.c_void_p(array.data.ptr), POINTER(dtype))
    return array, ptr


def _ptr_to_cupy(ptr, shape, dtype, owner, device):
    """
    Helper to convert a ctypes pointer to a CuPy array.
    """
    assert device.type == VesinCUDA, "Array is not on CUDA device"
    ptr_val = ctypes.cast(ptr, ctypes.c_void_p).value
    size = np.prod(shape) * ctypes.sizeof(dtype)
    mem = cp.cuda.memory.UnownedMemory(ptr_val, size, owner, device.device_id)
    return cp.ndarray(
        shape, dtype=CUPY_DTYPES[dtype], memptr=cp.cuda.MemoryPointer(mem, 0)
    )


########################################################################################


def _torch_asarray(array, device_like) -> "torch.Tensor":
    tensor = torch.as_tensor(array)
    if device_like is not None:
        return tensor.to(device=device_like.device)
    else:
        return tensor


def _torch_to_dtype(array: "torch.Tensor", dtype) -> "torch.Tensor":
    return array.to(dtype)


def _torch_empty(shape, dtype, device_like) -> "torch.Tensor":
    dtype = TORCH_DTYPES.get(dtype, dtype)
    return torch.empty(size=shape, dtype=dtype, device=device_like.device)


def _torch_get_ptr(array: "torch.Tensor", dtype) -> ("torch.Tensor", POINTER):
    array = array.contiguous()
    return array, ctypes.cast(ctypes.c_void_p(array.data_ptr()), POINTER(dtype))


CUDART = None


def _get_cudart():
    global CUDART
    if CUDART is not None:
        return CUDART

    import ctypes.util

    # try ctypes.util.find_library first
    cudart_path = ctypes.util.find_library("cudart")
    candidates = [cudart_path] if cudart_path else []

    # platform-specific fallbacks
    if sys.platform.startswith("win"):
        # common env vars and locations on Windows
        cuda_paths = []
        if "CUDA_PATH" in os.environ:
            cuda_paths.append(os.environ["CUDA_PATH"])
        if "CUDA_HOME" in os.environ:
            cuda_paths.append(os.environ["CUDA_HOME"])
        # look for typical cudart DLL names under bin/lib
        for cp in cuda_paths:
            candidates += glob.glob(os.path.join(cp, "bin", "cudart64_*.dll"))
            candidates += glob.glob(os.path.join(cp, "lib", "x64", "cudart64_*.dll"))
        # search directories on PATH
        for p in os.environ.get("PATH", "").split(os.pathsep):
            candidates += glob.glob(os.path.join(p, "cudart64_*.dll"))
    else:
        # linux: common install locations
        candidates += glob.glob("/usr/local/cuda*/lib64/libcudart.so*")
        candidates += glob.glob("/usr/lib*/cuda*/lib64/libcudart.so*")

    # unique, non-empty
    candidates = [c for c in set(candidates) if c and os.path.isfile(c)]

    last_err = None
    for path in candidates:
        try:
            CUDART = ctypes.CDLL(path)
            break
        except Exception as e:
            last_err = e
            continue

    if CUDART is None:
        raise RuntimeError(
            "Could not find cudart library, please make sure it is available or "
            "install cupy. On Windows ensure CUDA_PATH is set or cudart64_*.dll is "
            "on PATH."
        ) from last_err

    CUDART.cudaMemcpy.argtypes = [
        ctypes.c_void_p,
        ctypes.c_void_p,
        ctypes.c_size_t,
        ctypes.c_int,
    ]
    CUDART.cudaMemcpy.restype = ctypes.c_int
    return CUDART


def _ptr_to_torch(ptr, shape, dtype, owner, device):
    """
    Helper to convert a ctypes pointer to a Torch tensor.
    """

    if device.type == VesinCPU:
        array = _ptr_to_numpy(ptr, shape, dtype, owner, device)
        return torch.from_numpy(array)
    elif device.type == VesinCUDA:
        if HAS_CUPY:
            cupy_array = _ptr_to_cupy(ptr, shape, dtype, owner, device)
            return torch.as_tensor(cupy_array)
        else:
            cudart = _get_cudart()

            # TODO: we could try using from_dlpack to avoid a copy
            device = torch.device(f"cuda:{device.device_id}")
            tensor = torch.empty(size=shape, dtype=TORCH_DTYPES[dtype], device=device)
            tensor_ptr = ctypes.c_void_p(tensor.data_ptr())
            status = cudart.cudaMemcpy(
                tensor_ptr,
                ptr,
                tensor.nbytes,
                2,  # cudaMemcpyDeviceToDevice
            )
            if status != 0:
                raise RuntimeError(
                    f"cudaMemcpy failed when creating torch tensor (code {status})"
                )
            return tensor
    else:
        raise RuntimeError("Unknown device for torch tensor")
